use std::collections::BTreeSet;
use std::mem::take;

use crate::*;

/*
 * Top level function to definitively place functions onto devices. A function
 * may store a device placement, but only optionally - this function assigns
 * devices to the rest of the functions.
 */
pub fn device_placement(functions: &Vec<Function>, callgraph: &CallGraph) -> Vec<Device> {
    let mut devices = vec![];

    for (idx, function) in functions.into_iter().enumerate() {
        if let Some(device) = function.device {
            devices.push(device);
        } else if function.entry || callgraph.num_callees(FunctionID::new(idx)) != 0 {
            devices.push(Device::AsyncRust);
        } else {
            devices.push(Device::LLVM);
        }
    }

    devices
}

pub type FunctionObjectDeviceDemands = Vec<BTreeSet<Device>>;
pub type ObjectDeviceDemands = Vec<FunctionObjectDeviceDemands>;

/*
 * This analysis figures out which device each collection object may be on. At
 * first, an object may need to be on different devices at different times. This
 * is fine during optimization.
 */
pub fn object_device_demands(
    functions: &Vec<Function>,
    types: &Vec<Type>,
    typing: &ModuleTyping,
    callgraph: &CallGraph,
    objects: &CollectionObjects,
    devices: &Vec<Device>,
) -> ObjectDeviceDemands {
    // An object is "demanded" on a device when:
    // 1. The object is used by a primitive read node or write node in a device
    //    function. This includes objects on the `data` input to write nodes.
    //    Non-primitive reads don't demand an object on a device since they are
    //    lowered to pointer math and no actual memory transfers.
    // 2. The object is a constant / undef defined in a device function.
    // 3. The object is passed as input to a call node where the corresponding
    //    object in the callee is demanded on a device.
    // 4. The object is returned from a call node where the corresponding object
    //    in the callee is demanded on a device.
    // Note that reads and writes in a RT function don't induce a device demand.
    // This is because RT functions can  call device functions as necessary to
    // arbitrarily move data onto / off of devices (though this may be slow).
    // Traverse the functions in a module in reverse topological order, since
    // the analysis of a function depends on all functions it calls.
    let mut demands: ObjectDeviceDemands = vec![vec![]; functions.len()];
    let topo = callgraph.topo();

    for func_id in topo {
        let function = &functions[func_id.idx()];
        let typing = &typing[func_id.idx()];
        let device = devices[func_id.idx()];

        demands[func_id.idx()].resize(objects[&func_id].num_objects(), BTreeSet::new());
        match device {
            Device::LLVM | Device::CUDA => {
                for (idx, node) in function.nodes.iter().enumerate() {
                    match node {
                        // Condition #1.
                        Node::Read {
                            collect,
                            indices: _,
                        } if types[typing[idx].idx()].is_primitive() => {
                            for object in objects[&func_id].objects(*collect) {
                                demands[func_id.idx()][object.idx()].insert(device);
                            }
                        }
                        Node::Write {
                            collect,
                            data,
                            indices: _,
                        } => {
                            for object in objects[&func_id]
                                .objects(*collect)
                                .into_iter()
                                .chain(objects[&func_id].objects(*data).into_iter())
                            {
                                demands[func_id.idx()][object.idx()].insert(device);
                            }
                        }
                        // Condition #2.
                        Node::Constant { id: _ } | Node::Undef { ty: _ } => {
                            for object in objects[&func_id].objects(NodeID::new(idx)) {
                                demands[func_id.idx()][object.idx()].insert(device);
                            }
                        }
                        _ => {}
                    }
                }
            }
            Device::AsyncRust => {
                for (idx, node) in function.nodes.iter().enumerate() {
                    if let Node::Call {
                        control: _,
                        function: callee,
                        dynamic_constants: _,
                        args,
                    } = node
                    {
                        // Condition #3.
                        for (param_idx, arg) in args.into_iter().enumerate() {
                            if let Some(callee_obj) = objects[callee].param_to_object(param_idx) {
                                let callee_demands =
                                    take(&mut demands[callee.idx()][callee_obj.idx()]);
                                for object in objects[&func_id].objects(*arg) {
                                    demands[func_id.idx()][object.idx()]
                                        .extend(callee_demands.iter());
                                }
                                demands[callee.idx()][callee_obj.idx()] = callee_demands;
                            }
                        }

                        // Condition #4.
                        for callee_obj in objects[callee].returned_objects() {
                            let callee_demands = take(&mut demands[callee.idx()][callee_obj.idx()]);
                            for object in objects[&func_id].objects(NodeID::new(idx)) {
                                demands[func_id.idx()][object.idx()].extend(callee_demands.iter());
                            }
                            demands[callee.idx()][callee_obj.idx()] = callee_demands;
                        }
                    }
                }
            }
        }
    }

    demands
}
